==> wimlib-0.4.8.tar.gz: wimlib-0.4.8/src/xpress-decomp.c <==
/*
 * xpress-decomp.c
 *
 * XPRESS decompression routines.
 *
 * wimlib - Library for working with WIM files 
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option) any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */


/*
 * The XPRESS compression format is a LZ77-based algorithm.  That means it is
 * quite similar to LZX compression, but XPRESS is slightly simpler, so it is a
 * little faster to compress and decompress.
 *
 * The XPRESS compression format is mostly documented in a file called "[MS-XCA]
 * Xpress Compression Algorithm".  In the MSDN library, it can currently be
 * found under Open Specifications => Protocols => Windows Protocols => Windows
 * Server Protocols => [MS-XCA] Xpress Compression Algorithm".  Note that
 * Microsoft apparently also has either a slightly different format or an
 * entirely different format that is also called XPRESS.  The other one is
 * supposedly used in Windows' hibernation file or something, but the one used
 * in WIM files is the one described in the above document.
 *
 * If you are already familiar with the LZ77 algorithm and Huffman coding, the
 * XPRESS format is pretty simple.  The compressed data begins with 256 bytes
 * that contain 256 4-bit integers that are the lengths of the symbols in the
 * Huffman tree used for decoding compressed literals.  This is the only Huffman
 * tree that is used for the entirety of the compressed data, and the codeword
 * lengths are not encoded with a pretree.
 *
 * The rest of the compressed data is Huffman-encoded symbols.  Values 0 to 255
 * are literal bytes.  Values 256-512 are matches and may require extra bits or
 * bytes to be read to get the match offset and match length.
 *
 * There is no notion of a "compressed block" in the XPRESS format, so in the
 * XPRESS format, each WIM chunk (32768) bytes will always use only one Huffman
 * tree.
 */

#include "util.h"
#include "xpress.h"

#define XPRESS_DECOMP
#include "decomp.h"

#include "huffman.h"



/* Decodes @huffsym, a value >= XPRESS_NUM_CHARS, that is the header of a match.
 * */
static int
xpress_decode_match(int huffsym, unsigned window_pos, unsigned window_len,
			u8 window[], struct input_bitstream *istream)
{
	int match_len, match_offset;

	u8 match_sym = (u8)huffsym;
	u8 len_hdr = match_sym & 0xf;
	u8 offset_bsr = match_sym >> 4;

	if (!bitstream_read_bits(istream, offset_bsr, &match_offset))
		return -1;
	match_offset |= (1 << offset_bsr);

	/*XPRESS_DEBUG("huffsym = 0x%x, pos=%u, offset_bsr = %u, len_hdr = %u\n", */
				/*huffsym, window_pos,*/
			/*offset_bsr, len_hdr);*/

	if (len_hdr == 0xf) {
		int ret;
		ret = bitstream_read_byte(istream);
		if (ret == -1)
			return false;
		match_len = ret;
		if (match_len == 0xff) {

			ret = bitstream_read_byte(istream);
			if (ret == -1)
				return false;
			match_len = ret;

			ret = bitstream_read_byte(istream);
			if (ret == -1)
				return false;

			match_len |= (ret << 8);
			if (match_len < 0xf)
				return false;
		} else {
			match_len += 0xf;
		}
	} else {
		match_len = len_hdr;
	}
	match_len += XPRESS_MIN_MATCH_LEN;


	/* Verify that the match is in the bounds of the part of the window
	 * currently in use, then copy the source of the match to the current
	 * position. */

	u8 *match_dest = window + window_pos;
	u8 *match_src = match_dest - match_offset;

	if (window_pos + match_len > window_len) {
		wimlib_error("XPRESS dedecompression error: match of length %d "
				"bytes overflows window\n", match_len);
		return -1;
	}

	if (match_src < window) {
		wimlib_error("XPRESS decompression error: match of length %d bytes "
				"references data before window (match_offset = "
				"%d, window_pos = %d)\n", match_len,
				match_offset, window_pos);
		return -1;
	}

	for (int i = 0; i < match_len; i++)
		match_dest[i] = match_src[i];

	return match_len;
}

/* Decodes the Huffman-encoded matches and literal bytes in a block of
 * XPRESS-encoded data. */
static bool 
xpress_decompress_literals(struct input_bitstream *istream, 
				u8 uncompressed_data[], 
				unsigned uncompressed_len, 
				const u8 lens[], const u16 decode_table[])
{
	unsigned curpos = 0;
	while (curpos < uncompressed_len) {
		int huffsym;
		if (!read_huffsym(istream, decode_table, lens, 
				XPRESS_NUM_SYMBOLS, XPRESS_TABLEBITS, &huffsym,
				XPRESS_MAX_CODEWORD_LEN))
			return false;
		if (huffsym < XPRESS_NUM_CHARS) {
			uncompressed_data[curpos++] = huffsym;
		} else {
			int match_len = xpress_decode_match(huffsym, curpos, 
						uncompressed_len, 
						uncompressed_data, istream);
			if (match_len == -1)
				return false;
			curpos += match_len;
		}
	}
	return true;
}


bool xpress_decompress(const void *__compressed_data, int compressed_len, 
			void *uncompressed_data, int uncompressed_len)
{
	DEBUG2("compressed_len = %d, uncompressed_len = %d\n",
			compressed_len, uncompressed_len);
	const u8 *compressed_data = __compressed_data;

	/* XPRESS uses only one Huffman tree.  It contains 512 symbols, and the
	 * code lengths of these symbols are given literally as 4-bit integers
	 * in the first 256 bytes of the compressed data.
	 */
	if (compressed_len < XPRESS_NUM_SYMBOLS / 2)
		return false;

	u8 lens[XPRESS_NUM_SYMBOLS];
	u16 decode_table[(1 << XPRESS_TABLEBITS) + 2 * XPRESS_NUM_SYMBOLS];

	u8 *lens_p = lens;
	for (unsigned i = 0; i < XPRESS_NUM_SYMBOLS / 2; i++) {
		*lens_p++ = compressed_data[i] & 0xf;
		*lens_p++ = compressed_data[i] >> 4;
	}

	if (!make_huffman_decode_table(decode_table, XPRESS_NUM_SYMBOLS,
					XPRESS_TABLEBITS, lens,
					XPRESS_MAX_CODEWORD_LEN))
		return false;

	struct input_bitstream istream;
	init_input_bitstream(&istream, compressed_data + XPRESS_NUM_SYMBOLS / 2, 
					compressed_len - XPRESS_NUM_SYMBOLS / 2);

	if (!xpress_decompress_literals(&istream, uncompressed_data, 
				uncompressed_len, lens, decode_table))
		return false;
	return true;
}

==> wimlib-0.5.0.tar.gz: wimlib-0.5.0/src/xpress-decomp.c <==
/*
 * xpress-decomp.c
 *
 * XPRESS decompression routines.
 *
 * Copyright (C) 2012 Eric Biggers
 *
 * wimlib - Library for working with WIM files 
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option) any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */



/*
 * The XPRESS compression format is a LZ77-based algorithm.  That means it is
 * quite similar to LZX compression, but XPRESS is slightly simpler, so it is a
 * little faster to compress and decompress.
 *
 * The XPRESS compression format is mostly documented in a file called "[MS-XCA]
 * Xpress Compression Algorithm".  In the MSDN library, it can currently be
 * found under Open Specifications => Protocols => Windows Protocols => Windows
 * Server Protocols => [MS-XCA] Xpress Compression Algorithm".  Note that
 * Microsoft apparently also has either a slightly different format or an
 * entirely different format that is also called XPRESS.  The other one is
 * supposedly used in Windows' hibernation file or something, but the one used
 * in WIM files is the one described in the above document.
 *
 * If you are already familiar with the LZ77 algorithm and Huffman coding, the
 * XPRESS format is pretty simple.  The compressed data begins with 256 bytes
 * that contain 512 4-bit integers that are the lengths of the symbols in the
 * Huffman tree used for decoding compressed literals.  This is the only Huffman
 * tree that is used for the entirety of the compressed data, and the codeword
 * lengths are not encoded with a pretree.
 *
 * The rest of the compressed data is Huffman-encoded symbols.  Values 0 through
 * 255 are literal bytes.  Values 256 through 511 are matches and may require
 * extra bits or bytes to be read to get the match offset and match length.
 *
 * There is no notion of a "compressed block" in the XPRESS format, so in the
 * XPRESS format, each WIM chunk (32768 bytes) will always use only one Huffman
 * tree.
 *
 * The trickiest part is probably the fact that literal bytes for match lengths
 * are encoded "separately" from the bitstream.
 *
 * Also, a caveat--- according to M$'s documentation for XPRESS,
 *
 * 	"Some implementation of the decompression algorithm expect an extra
 * 	symbol to mark the end of the data.  Specifically, some implementations
 * 	fail during decompression if the Huffman symbol 256 is not found after
 * 	the actual data."
 *
 * This is the case for WIM files--- in we must write this extra symbol "256" at
 * the end.  Otherwise Microsoft's software will fail to decompress the
 * XPRESS-compressed data.
 *
 * Howeve, WIMLIB's decompressor in xpress-decomp.c currently does not care if
 * this extra symbol is there or not.
 */

#include "util.h"
#include "xpress.h"

#define XPRESS_DECOMP
#include "decomp.h"

#include "huffman.h"



/* Decodes @huffsym, a value >= XPRESS_NUM_CHARS, that is the header of a match.
 * */
static int xpress_decode_match(int huffsym, uint window_pos, uint window_len, 
				u8 window[], struct input_bitstream *istream)
{
	int match_len, match_offset;

	u8 match_sym = (u8)huffsym;
	u8 len_hdr = match_sym & 0xf;
	u8 offset_bsr = match_sym >> 4;

	if (!bitstream_read_bits(istream, offset_bsr, &match_offset))
		return -1;
	match_offset |= (1 << offset_bsr);

	if (len_hdr == 0xf) {
		int ret;
		ret = bitstream_read_byte(istream);
		if (ret == -1)
			return false;
		match_len = ret;
		if (match_len == 0xff) {

			ret = bitstream_read_byte(istream);
			if (ret == -1)
				return false;
			match_len = ret;

			ret = bitstream_read_byte(istream);
			if (ret == -1)
				return false;

			match_len |= (ret << 8);
			if (match_len < 0xf)
				return false;
		} else {
			match_len += 0xf;
		}
	} else {
		match_len = len_hdr;
	}
	match_len += XPRESS_MIN_MATCH;


	/*XPRESS_DEBUG("match_len = %u, match_offset = %u\n",*/
			/*match_len, match_offset);*/

	/* Verify that the match is in the bounds of the part of the window
	 * currently in use, then copy the source of the match to the current
	 * position. */

	u8 *match_dest = window + window_pos;
	u8 *match_src = match_dest - match_offset;

	if (window_pos + match_len > window_len) {
		wimlib_error("XPRESS dedecompression error: match of length %d "
				"bytes overflows window\n", match_len);
		return -1;
	}

	if (match_src < window) {
		wimlib_error("XPRESS decompression error: match of length %d bytes "
				"references data before window (match_offset = "
				"%d, window_pos = %d)\n", match_len,
				match_offset, window_pos);
		return -1;
	}

	for (int i = 0; i < match_len; i++)
		match_dest[i] = match_src[i];

	return match_len;
}

/* Decodes the Huffman-encoded matches and literal bytes in a block of
 * XPRESS-encoded data. */
static bool xpress_decompress_literals(struct input_bitstream *istream, 
				u8 uncompressed_data[], uint uncompressed_len, 
				const u8 lens[], const u16 decode_table[])
{
	uint curpos = 0;
	while (curpos < uncompressed_len) {
		int huffsym;
		if (!read_huffsym(istream, decode_table, lens, 
				XPRESS_NUM_SYMBOLS, XPRESS_TABLEBITS, &huffsym,
				XPRESS_MAX_CODEWORD_LEN))
			return false;
		if (huffsym < XPRESS_NUM_CHARS) {
			/*XPRESS_DEBUG("decoded literal (pos = %u, value = %u)\n", */
						/*curpos, huffsym);*/
			uncompressed_data[curpos++] = huffsym;
		} else {
			int match_len = xpress_decode_match(huffsym, curpos, 
						uncompressed_len, 
						uncompressed_data, istream);
			/*XPRESS_DEBUG("decoded match (pos = %u, len = %u)\n",*/
					/*curpos, match_len);*/
			if (match_len == -1)
				return false;
			curpos += match_len;
		}
	}
	return true;
}


bool xpress_decompress(const void *__compressed_data, int compressed_len, 
			void *uncompressed_data, int uncompressed_len)
{
	DEBUG2("compressed_len = %d, uncompressed_len = %d\n",
			compressed_len, uncompressed_len);
	const u8 *compressed_data = __compressed_data;

	/* XPRESS uses only one Huffman tree.  It contains 512 symbols, and the
	 * code lengths of these symbols are given literally as 4-bit integers
	 * in the first 256 bytes of the compressed data.
	 */
	if (compressed_len < XPRESS_NUM_SYMBOLS / 2)
		return false;

	u8 lens[XPRESS_NUM_SYMBOLS];
	u16 decode_table[(1 << XPRESS_TABLEBITS) + 2 * XPRESS_NUM_SYMBOLS];

	u8 *lens_p = lens;
	for (uint i = 0; i < XPRESS_NUM_SYMBOLS / 2; i++) {
		*lens_p++ = compressed_data[i] & 0xf;
		*lens_p++ = compressed_data[i] >> 4;
	}

	if (!make_huffman_decode_table(decode_table, XPRESS_NUM_SYMBOLS,
					XPRESS_TABLEBITS, lens,
					XPRESS_MAX_CODEWORD_LEN))
		return false;

	struct input_bitstream istream;
	init_input_bitstream(&istream, compressed_data + XPRESS_NUM_SYMBOLS / 2, 
					compressed_len - XPRESS_NUM_SYMBOLS / 2);

	if (!xpress_decompress_literals(&istream, uncompressed_data, 
				uncompressed_len, lens, decode_table))
		return false;
	return true;
}
